#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <cuda.h>
#include <sys/time.h>
#include <math.h>

#define CHK_CUDA(e) {if (e != cudaSuccess) {fprintf(stderr,"Error: %s\n", cudaGetErrorString(e)); exit(-1);}}
#define CUDA_ERRCHECK() { cudaError e=cudaGetLastError(); CHK_CUDA(e); }

__global__ 
void cuda_sgemm_simple(const float *ad, const float *bd, const int rowsA, const int rowsB, const int colsB, float *cd) {

	int i = threadIdx.y + blockIdx.y*blockDim.y;
	int j = threadIdx.x + blockIdx.x*blockDim.x;
    int k;
    const int colsA=rowsB;
    double sum;
	if (i<rowsA && j<colsB) {
        sum = 0.;
        for (k=0; k<rowsB; k++) {
            sum += ad[i*colsA + k]*bd[k*colsB + j];
        }
        cd[i*colsB + j] = sum;
	}
	return;
}

__global__ void noop(const float *ad, const float *bd, const int rowsA, const int rowsB, const int colsB, float *cd) {
    return;
}



void cpu_sgemm_simple(const float *a, const float *b, const int rowsA, const int rowsB, const int colsB, float *c) {

    /* this, of course, is a  
        terrible implementation */

	int i, j, k;
	for (i=0; i<rowsA; i++) {
        for (j=0; j<colsB; j++) {
            double sum = 0.;
            for (k=0; k<rowsB; k++) {
                sum += a[i*rowsB + k]*b[k*colsB + j];
            }
            c[i*colsB + j] = sum;
		}
	}
	return;
}

int get_options(int argc, char **argv, int *n, int *nb, long int *seed);
void tick(struct timeval *timer);
double tock(struct timeval *timer);
void printmatrix(const char *hdr, const float *a, const int n);

int main(int argc, char **argv) {
	int n=160;
	int nblocks=10;
	int err;
	float *a, *b, *c, *ccuda;
	float *ad, *bd, *cd;
    float *cdiff;
	double abserr;
	dim3 blocksize, gridsize;
    long int seed;
    int i;
    struct timeval cputimer;
    struct timeval gputimer;
    double cputime, gputime;

	err = get_options(argc, argv, &n, &nblocks, &seed);
	if (err) return 0;
    
    if ((n % nblocks) != 0) {
        fprintf(stderr,"Error: Number of blocks %d must evenly divide matrix size %d.\n",
                 nblocks, n);
        return -1;
    }

    srand48(seed);

    cudaHostAlloc(&a, n*n*sizeof(float), cudaHostAllocDefault);
    cudaHostAlloc(&b, n*n*sizeof(float), cudaHostAllocDefault);
    cudaHostAlloc(&c, n*n*sizeof(float), cudaHostAllocDefault);
    cudaHostAlloc(&ccuda, n*n*sizeof(float), cudaHostAllocDefault);

	cdiff = (float *)malloc(n*n*sizeof(float));
	if (!cdiff) {free(ccuda); free(c); free(b); free(a); return 1;}

    for (i=0; i < n*n; i++) {
        a[i] = (float)drand48();
        b[i] = (float)drand48();
    }

	/* run CPU code */

    tick(&cputimer);
    cpu_sgemm_simple(a, b, n, n, n, c);
    cputime = tock(&cputimer);

	/* Dummy kernel for warm start*/
	CHK_CUDA( cudaMalloc(&ad, sizeof(float)) );
	CHK_CUDA( cudaMalloc(&bd, sizeof(float)) );
	CHK_CUDA( cudaMalloc(&cd, sizeof(float)) );

    tick(&gputimer);
	CHK_CUDA( cudaMemcpy(ad, a, sizeof(float), cudaMemcpyHostToDevice) );
	CHK_CUDA( cudaMemcpy(bd, b, sizeof(float), cudaMemcpyHostToDevice) );

    noop<<<100,100>>>(ad, bd, n, n, n, cd);
    CUDA_ERRCHECK();

	CHK_CUDA( cudaMemcpy(ccuda, cd, sizeof(float), cudaMemcpyDeviceToHost) );
    gputime = tock(&gputimer);
	
	CHK_CUDA( cudaFree(ad) );
	CHK_CUDA( cudaFree(bd) );
	CHK_CUDA( cudaFree(cd) );


	/* run GPU code */
	CHK_CUDA( cudaMalloc(&ad, n*n*sizeof(float)) );
	CHK_CUDA( cudaMalloc(&bd, n*n*sizeof(float)) );
	CHK_CUDA( cudaMalloc(&cd, n*n*sizeof(float)) );

    tick(&gputimer);
	CHK_CUDA( cudaMemcpy(ad, a, n*n*sizeof(float), cudaMemcpyHostToDevice) );
	CHK_CUDA( cudaMemcpy(bd, b, n*n*sizeof(float), cudaMemcpyHostToDevice) );

	blocksize = make_uint3( (n+nblocks-1)/nblocks, (n+nblocks-1)/nblocks, 1);
    gridsize  = make_uint3( nblocks, nblocks, 1);

    cuda_sgemm_simple<<<gridsize, blocksize>>>(ad, bd, n, n, n, cd);
    CUDA_ERRCHECK();

	CHK_CUDA( cudaMemcpy(ccuda, cd, n*n*sizeof(float), cudaMemcpyDeviceToHost) );
    gputime = tock(&gputimer);
	
	CHK_CUDA( cudaFree(ad) );
	CHK_CUDA( cudaFree(bd) );
	CHK_CUDA( cudaFree(cd) );

    abserr = 0.;
    for (i=0;i<n*n;i++) {
        cdiff[i] = fabs(ccuda[i] - c[i]);
        abserr += cdiff[i];
    }
    
    if (n<20) printmatrix("Difference", cdiff, n);
    
    printf("Matrix size = %d, Number of blocks = %d.\n", n, nblocks);
    printf("CPU time = %lg millisec.\n", cputime*1000.);
    printf("GPU time = %lg millisec (1 gpu).\n", gputime*1000.);
    printf("Single GPU and CPU results differ by %lf\n", abserr);


    /* run on as many gpus as possible */

    int ngpus;
	CHK_CUDA( cudaGetDeviceCount( &ngpus ));
    
    int nrows     = n/ngpus;
    int lastnrows = n - nrows*(ngpus-1);

    float **multi_ad = (float **)malloc((ngpus)*sizeof(float **));
    float **multi_bd = (float **)malloc((ngpus)*sizeof(float **));
    float **multi_cd = (float **)malloc((ngpus)*sizeof(float **));

    cudaStream_t *streams = (cudaStream_t *)malloc((ngpus)*sizeof(cudaStream_t));

    for (int dev=0; dev<ngpus; dev++) {
        cudaSetDevice(dev);

        cudaStreamCreate(&(streams[dev]));

        int locnrows = nrows;
        if (dev == ngpus-1) locnrows = lastnrows;

        CHK_CUDA( cudaMalloc(&(multi_ad[dev]), locnrows*n*sizeof(float)) );
        CHK_CUDA( cudaMalloc(&(multi_bd[dev]), n*n*sizeof(float)) );
        CHK_CUDA( cudaMalloc(&(multi_cd[dev]), locnrows*n*sizeof(float)) );
    }

    tick(&gputimer);

    for (int dev=0; dev<ngpus; dev++) {
        cudaSetDevice(dev);

        int locnrows = nrows;
        if (dev == ngpus-1) locnrows = lastnrows;

        CHK_CUDA( cudaMemcpyAsync(multi_ad[dev], &(a[n*(nrows*dev)]), locnrows*n*sizeof(float), cudaMemcpyHostToDevice, streams[dev]) );
        CHK_CUDA( cudaMemcpyAsync(multi_bd[dev], b, n*n*sizeof(float), cudaMemcpyHostToDevice, streams[dev]) );
    }


    for (int dev=0; dev<ngpus; dev++) {
        cudaSetDevice(dev);

        int locnrows = nrows;
        if (dev == ngpus-1) locnrows = lastnrows;

        int gridrows = locnrows/blocksize.y;
        int gridcols = n/blocksize.x;
        gridsize  = make_uint3( gridcols, gridrows, 1);

        cuda_sgemm_simple<<<gridsize, blocksize>>>(multi_ad[dev], multi_bd[dev], locnrows, n, n, multi_cd[dev]);
        CUDA_ERRCHECK();
    }


    for (int dev=0; dev<ngpus; dev++) {
        cudaSetDevice(dev);

        int locnrows = nrows;
        if (dev == ngpus-1) locnrows = lastnrows;

        CHK_CUDA( cudaMemcpyAsync(&(ccuda[nrows*dev*n]), multi_cd[dev], locnrows*n*sizeof(float), cudaMemcpyDeviceToHost, streams[dev]) );
    }

    for (int dev=0; dev<ngpus; dev++) {
        cudaSetDevice(dev);
        cudaDeviceSynchronize();
    }

    gputime = tock(&gputimer);

    for (int dev=0; dev<ngpus; dev++) {
        cudaSetDevice(dev);

        CHK_CUDA( cudaFree(multi_ad[dev]) );
        CHK_CUDA( cudaFree(multi_bd[dev]) );
        CHK_CUDA( cudaFree(multi_cd[dev]) );
    }

    abserr = 0.;
    for (i=0;i<n*n;i++) {
        cdiff[i] = fabs(ccuda[i] - c[i]);
        abserr += cdiff[i];
    }
    
    if (n<20) printmatrix("Difference", cdiff, n);
    
    printf("GPU time = %lg millisec (%d gpu).\n", ngpus, gputime*1000.);
    printf("%d-GPU and CPU results differ by %lf\n", ngpus, abserr);

    cudaFreeHost(a);
    cudaFreeHost(b);
    cudaFreeHost(c);
    cudaFreeHost(ccuda);
    free(cdiff);
	return 0;
}


int get_options(int argc, char **argv, int *n, int *nb, long int *seed) {

  const struct option long_options[] = {
    {"matsize" , required_argument, 0, 'n'},
    {"nblocks" , required_argument, 0, 'B'},
    {"seed"    , required_argument, 0, 's'},
    {"help",      no_argument, 0, 'h'},
    {0, 0, 0, 0}};

  char c;
  int option_index;
  int tempint;

  while (1) {
    c = getopt_long(argc, argv, "n:B:a:b:h", long_options, &option_index);
    if (c == -1) break;
 
    switch(c) {
      case 'n': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 50000) {
            fprintf(stderr,"%s: Cannot use number of columns %s;\n  Using %d\n", argv[0], optarg, *n);
          } else {
            *n = tempint;
          }
          break;

      case 'B': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 1000 || tempint > *n) {
            fprintf(stderr,"%s: Cannot use number of blocks %s;\n  Using %d\n", argv[0], optarg, *nb);
          } else {
            *nb = tempint;
          }
          break;

      case 's': *seed = atoi(optarg);
          break;

      case 'h':
          puts("Calculates y[i] = a*x[i] + b on the GPU.");
          puts("Options: ");
          puts("    --matsize=N    (-n N): Set the number of values in y,x.");
          puts("    --nblocks=N    (-B N): Set the number of blocks used in each dimension.");
          puts("    --seed=N       (-s N): Seed the random number generator.");
          puts("");
          return +1;
        }
    }

    return 0;
}

void tick(struct timeval *timer) {
    gettimeofday(timer, NULL);
}

double tock(struct timeval *timer) {
    struct timeval now;
    gettimeofday(&now, NULL);
    return (now.tv_usec-timer->tv_usec)/1.0e6 + (now.tv_sec - timer->tv_sec);
}

void printmatrix(const char *hdr, const float *a, const int n) {
    int i,j;
    printf("%s\n", hdr);

    for (i=0; i<n; i++) {
        for (j=0; j<n; j++) {
            printf("%7.5g ", a[i*n+j]);
        } 
        puts("");
    }
}
