#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <cuda.h>
#include <sys/time.h>
#include <math.h>

#define CHK_CUDA(e) {if (e != cudaSuccess) {fprintf(stderr,"Error: %s\n", cudaGetErrorString(e)); exit(-1);}}

__global__ void cuda_transpose_shared_multiple(const float *ad, const int n, const int multi, float *atd) {

    extern __shared__ float atile[];
    
	int loci = threadIdx.y;
	int locj = threadIdx.x;
    int bi = blockIdx.y;
    int bj = blockIdx.x;
	int i = loci + bi*blockDim.y;
	int j = locj + bj*blockDim.x;

    /* read in multipel rows into shared data */
    if (j<n) {
        for (int iter=0; iter<multi; iter++) {
            int inrow  =multi*i+iter;
            int tilerow=multi*loci+iter;
            if ( inrow < n) {
                atile[tilerow*(blockDim.x+1) + locj] = ad[inrow*n + j];
            }
        }
    }

    __syncthreads();

    for (int iter=0; iter<multi; iter++) { 
        int outrow=blockDim.x*bj + loci/multi;
        int outcol=blockDim.y*bi*multi + (loci%multi)*blockDim.y + locj;
            
        if (outrow < n && outcol < n) {
            atd[outrow*n + outcol] = atile[(multi*locj+iter)*(blockDim.x+1)+loci];
        }
    }
	return;
}

__global__ void cuda_transpose_shared_bank(const float *ad, const int n, float *atd) {

    extern __shared__ float atile[];
    
	int loci = threadIdx.y;
	int locj = threadIdx.x;
    int bi = blockIdx.y;
    int bj = blockIdx.x;
    int blocksize = blockDim.x;
	int i = loci + bi*blocksize;
	int j = locj + bj*blocksize;

    /* read in shared data */
	if (i<n && j<n) {
        atile[loci*(blocksize+1) + locj] = ad[i*n + j];
    }
    __syncthreads();

    if (i<n && j<n) {
        atd[(loci+bj*blocksize)*n + (locj+bi*blocksize)] = atile[locj*(blocksize+1)+loci];
    }
	return;
}


__global__ void cuda_transpose_shared(const float *ad, const int n, float *atd) {

    extern __shared__ float atile[];
    
	int loci = threadIdx.y;
	int locj = threadIdx.x;
    int bi = blockIdx.y;
    int bj = blockIdx.x;
    int blocksize = blockDim.x;
	int i = loci + bi*blocksize;
	int j = locj + bj*blocksize;

    /* read in shared data */
	if (i<n && j<n) {
        atile[loci*blocksize + locj] = ad[i*n + j];
    }
    __syncthreads();

    if (i<n && j<n) {
        atd[(loci+bj*blocksize)*n + (locj+bi*blocksize)] = atile[locj*blocksize+loci];
    }
	return;
}

__global__ void cuda_transpose(const float *ad, const int n, float *atd) {

	int i = threadIdx.y + blockIdx.y*blockDim.y;
	int j = threadIdx.x + blockIdx.x*blockDim.x;
	if (i<n && j<n) 
        atd[i*n + j] = ad[j*n + i];
	return;
}

void cpu_transpose(const float *a, const int n, float *at) {
	for (int i=0;i<n;i++) {
        for (int j=0;j<n;j++) {
            at[i*n + j] = a[j*n+i];
		}
	}
	return;
}

int get_options(int argc, char **argv, int *n, int *nb, int *multi, long int *seed);
void tick(struct timeval *timer);
double tock(struct timeval *timer);

int main(int argc, char **argv) {
	int n=160;
	int nblocks=10;
    int multi=4;
	int err;
	float *a, *at, *atcuda;
	float *ad, *atd;
    float *cdiff;
	double abserr;
	dim3 blocksize, gridsize;
    long int seed;
    int i;
    struct timeval cputimer;
    struct timeval gputimer;
    double cputime, gputime;

	err = get_options(argc, argv, &n, &nblocks, &multi, &seed);
	if (err) return 0;
    
    if ((n % nblocks) != 0) {
        fprintf(stderr,"Error: Number of blocks %d must evenly divide matrix size %d.\n",
                 nblocks, n);
        return -1;
    }

    srand48(seed);

	a = (float *)malloc(n*n*sizeof(float));
	if (!a) return 1;
		
	at = (float *)malloc(n*n*sizeof(float));
	if (!at) {free(a); return 1;}

	atcuda = (float *)malloc(n*n*sizeof(float));
	if (!atcuda) {free(at); free(a); return 1;}

	cdiff = (float *)malloc(n*n*sizeof(float));
	if (!cdiff) {free(atcuda); free(at); free(a); return 1;}

    for (i=0; i < n*n; i++) {
        a[i] = (float)drand48();
    }

	/* run CPU code */
    tick(&cputimer);
    cpu_transpose(a, n, at);
    cputime = tock(&cputimer);

	/* run GPU code once for cold start*/

	CHK_CUDA( cudaMalloc(&ad,  n*n*sizeof(float)) );
	CHK_CUDA( cudaMalloc(&atd, n*n*sizeof(float)) );

    tick(&gputimer);
	CHK_CUDA( cudaMemcpy(ad, a, n*n*sizeof(float), cudaMemcpyHostToDevice) );

	blocksize = make_uint3( (n+nblocks-1)/nblocks, (n+nblocks-1)/nblocks, 1);
    gridsize  = make_uint3( nblocks, nblocks, 1);

    cuda_transpose<<<gridsize, blocksize>>>(ad, n, atd);

	CHK_CUDA( cudaMemcpy(atcuda, atd, n*n*sizeof(float), cudaMemcpyDeviceToHost) );
    gputime = tock(&gputimer);
	
	CHK_CUDA( cudaFree(ad) );
	CHK_CUDA( cudaFree(atd) );

	/* run GPU code once for real */
	CHK_CUDA( cudaMalloc(&ad,  n*n*sizeof(float)) );
	CHK_CUDA( cudaMalloc(&atd, n*n*sizeof(float)) );

    tick(&gputimer);
	CHK_CUDA( cudaMemcpy(ad, a, n*n*sizeof(float), cudaMemcpyHostToDevice) );

	blocksize = make_uint3( (n+nblocks-1)/nblocks, (n+nblocks-1)/nblocks, 1);
    gridsize  = make_uint3( nblocks, nblocks, 1);

    cuda_transpose<<<gridsize, blocksize>>>(ad, n, atd);

	CHK_CUDA( cudaMemcpy(atcuda, atd, n*n*sizeof(float), cudaMemcpyDeviceToHost) );
    gputime = tock(&gputimer);
	
	CHK_CUDA( cudaFree(ad) );
	CHK_CUDA( cudaFree(atd) );

    abserr = 0.;
    for (i=0;i<n*n;i++) {
        cdiff[i] = fabs(atcuda[i] - at[i]);
        abserr += cdiff[i];
    }
    
    printf("Matrix size = %d, Number of blocks = %d, Blocksize = %d.\n", n, nblocks, blocksize.x);
    printf("CPU time = %lg millisec.\n", cputime*1000.);
    printf("GPU time = %lg millisec (global mem).\n", gputime*1000.);
    printf("CUDA global mem and CPU results differ by %lf\n", abserr);


	/* shmem GPU code */
	CHK_CUDA( cudaMalloc(&ad,  n*n*sizeof(float)) );
	CHK_CUDA( cudaMalloc(&atd, n*n*sizeof(float)) );

    tick(&gputimer);
	CHK_CUDA( cudaMemcpy(ad, a, n*n*sizeof(float), cudaMemcpyHostToDevice) );

    cuda_transpose_shared<<<gridsize, blocksize, 
                           blocksize.x*blocksize.y*sizeof(float)>>>(ad, n, atd);

	CHK_CUDA( cudaMemcpy(atcuda, atd, n*n*sizeof(float), cudaMemcpyDeviceToHost) );
    gputime = tock(&gputimer);
	
	CHK_CUDA( cudaFree(ad) );
	CHK_CUDA( cudaFree(atd) );

    abserr = 0.;
    for (i=0;i<n*n;i++) {
        cdiff[i] = fabs(atcuda[i] - at[i]);
        abserr += cdiff[i];
    }
    
    printf("GPU time = %lg millisec (shared mem).\n", gputime*1000.);
    printf("CUDA shared mem and CPU results differ by %lf\n", abserr);

	/* shmem GPU code + offset for shmem bank buffering */
	CHK_CUDA( cudaMalloc(&ad,  n*n*sizeof(float)) );
	CHK_CUDA( cudaMalloc(&atd, n*n*sizeof(float)) );

    tick(&gputimer);
	CHK_CUDA( cudaMemcpy(ad, a, n*n*sizeof(float), cudaMemcpyHostToDevice) );

    cuda_transpose_shared<<<gridsize, blocksize, 
                           (blocksize.x+1)*blocksize.y*sizeof(float)>>>(ad, n, atd);

	CHK_CUDA( cudaMemcpy(atcuda, atd, n*n*sizeof(float), cudaMemcpyDeviceToHost) );
    gputime = tock(&gputimer);
	
	CHK_CUDA( cudaFree(ad) );
	CHK_CUDA( cudaFree(atd) );

    abserr = 0.;
    for (i=0;i<n*n;i++) {
        cdiff[i] = fabs(atcuda[i] - at[i]);
        abserr += cdiff[i];
    }
    
    printf("GPU time = %lg millisec (shared mem).\n", gputime*1000.);
    printf("CUDA shared mem - bank and CPU results differ by %lf\n", abserr);


	/* shmem GPU code + offset for shmem bank buffering */
	CHK_CUDA( cudaMalloc(&ad,  n*n*sizeof(float)) );
	CHK_CUDA( cudaMalloc(&atd, n*n*sizeof(float)) );

    tick(&gputimer);
	CHK_CUDA( cudaMemcpy(ad, a, n*n*sizeof(float), cudaMemcpyHostToDevice) );

	blocksize = make_uint3( (n+nblocks-1)/nblocks, (n+nblocks*multi-1)/nblocks*multi, 1);
    gridsize  = make_uint3( nblocks, nblocks*multi, 1);

    cuda_transpose_shared_multiple<<<gridsize, blocksize, 
                           (blocksize.x+1)*blocksize.y*multi*sizeof(float)>>>(ad, n, multi, atd);

	CHK_CUDA( cudaMemcpy(atcuda, atd, n*n*sizeof(float), cudaMemcpyDeviceToHost) );
    gputime = tock(&gputimer);
	
	CHK_CUDA( cudaFree(ad) );
	CHK_CUDA( cudaFree(atd) );

    abserr = 0.;
    for (i=0;i<n*n;i++) {
        cdiff[i] = fabs(atcuda[i] - at[i]);
        abserr += cdiff[i];
    }
    
    printf("GPU time = %lg millisec (multiple transactions).\n", gputime*1000.);
    printf("CUDA multiple transactions and CPU results differ by %lf\n", abserr);

    free(a);
    free(at);
    free(atcuda);
    free(cdiff);
	return 0;
}


int get_options(int argc, char **argv, int *n, int *nb, int *multi, long int *seed) {

  const struct option long_options[] = {
    {"matsize" , required_argument, 0, 'n'},
    {"multi"   , required_argument, 0, 'm'},
    {"nblocks" , required_argument, 0, 'B'},
    {"seed"    , required_argument, 0, 's'},
    {"help",      no_argument, 0, 'h'},
    {0, 0, 0, 0}};

  char c;
  int option_index;
  int tempint;

  while (1) {
    c = getopt_long(argc, argv, "n:B:a:b:h", long_options, &option_index);
    if (c == -1) break;
 
    switch(c) {
      case 'n': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 50000) {
            fprintf(stderr,"%s: Cannot use number of columns %s;\n  Using %d\n", argv[0], optarg, *n);
          } else {
            *n = tempint;
          }
          break;

      case 'm': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 10) {
            fprintf(stderr,"%s: Cannot use transaction number %s;\n  Using %d\n", argv[0], optarg, *multi);
          } else {
            *multi = tempint;
          }
          break;

      case 'B': tempint = atoi(optarg);
          if (tempint < 1 || tempint > 1024 || tempint > *n) {
            fprintf(stderr,"%s: Cannot use number of blocks %s;\n  Using %d\n", argv[0], optarg, *nb);
          } else {
            *nb = tempint;
          }
          break;

      case 's': *seed = atoi(optarg);
          break;

      case 'h':
          puts("Calculates y[i] = a*x[i] + b on the GPU.");
          puts("Options: ");
          puts("    --matsize=N    (-n N): Set the number of values in y,x.");
          puts("    --nblocks=N    (-B N): Set the number of blocks used in each dimension.");
          puts("    --seed=N       (-s N): Seed the random number generator.");
          puts("");
          return +1;
        }
    }

    return 0;
}

void tick(struct timeval *timer) {
    gettimeofday(timer, NULL);
}

double tock(struct timeval *timer) {
    struct timeval now;
    gettimeofday(&now, NULL);
    return (now.tv_usec-timer->tv_usec)/1.0e6 + (now.tv_sec - timer->tv_sec);
}
